/*
SDX: Documentary System in XML.
Copyright (C) 2000, 2001, 2002  Ministere de la culture et de la communication (France), AJLSM

Ministere de la culture et de la communication,
Mission de la recherche et de la technologie
3 rue de Valois, 75042 Paris Cedex 01 (France)
mrt@culture.fr, michel.bottin@culture.fr

AJLSM, 17, rue Vital Carles, 33000 Bordeaux (France)
sevigny@ajlsm.com

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the
Free Software Foundation, Inc.
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
or connect to:
http://www.fsf.org/copyleft/gpl.html
 */
package fr.gouv.culture.sdx.documentbase;

// SDX imports

//import fr.gouv.culture.sdx.query.SDXResults;

import fr.gouv.culture.sdx.application.Application;
import fr.gouv.culture.sdx.document.*;
import fr.gouv.culture.sdx.exception.SDXException;
import fr.gouv.culture.sdx.exception.SDXExceptionCode;
import fr.gouv.culture.sdx.framework.Framework;
import fr.gouv.culture.sdx.framework.FrameworkImpl;
import fr.gouv.culture.sdx.pipeline.GenericPipeline;
import fr.gouv.culture.sdx.repository.Repository;
import fr.gouv.culture.sdx.repository.RepositoryConnection;
import fr.gouv.culture.sdx.repository.URLRepository;
import fr.gouv.culture.sdx.search.lucene.FieldsDefinition;
import fr.gouv.culture.sdx.search.lucene.analysis.MetaAnalyzer;
import fr.gouv.culture.sdx.search.lucene.query.LuceneIndex;
import fr.gouv.culture.sdx.utils.Utilities;
import fr.gouv.culture.sdx.utils.constants.Node;
import fr.gouv.culture.sdx.utils.database.DatabaseEntity;
import fr.gouv.culture.sdx.utils.database.LuceneDatabase;
import fr.gouv.culture.sdx.utils.database.Property;
import org.apache.avalon.excalibur.xml.Parser;
import org.apache.avalon.framework.component.ComponentException;
import org.apache.avalon.framework.configuration.Configuration;
import org.apache.avalon.framework.configuration.ConfigurationException;
import org.apache.avalon.framework.parameters.Parameters;
import org.apache.cocoon.ProcessingException;
import org.apache.cocoon.xml.XMLConsumer;
import org.apache.cocoon.xml.XMLizable;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.Term;
import org.apache.lucene.search.Hits;
import org.apache.lucene.search.TermQuery;
import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.DateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.Hashtable;

/**
 *	A document base within an SDX application.
 *
 * <p>
 * A document base is a very important concept in SDX development. A document base
 * is where documents are searched and retrieved, thus added (indexed), deleted or
 * updated. A search cannot occur in a smaller unit than the document base. To exclude
 * some parts of a document base, one should use query constructions, possibly filters.
 * <p>
 * A document base has a structure ; this structure is basically a list of fields. An
 * application may have many document bases, and these document bases may have different
 * structures. As always, indexable documents (XML, HTML or the like) with different
 * structures can be indexed within a single document base.
 * <p>
 * Most applications will have only one document base, but in some cases it could
 * be interesting to have more than one, like when different kinds of documents are
 * never searched at the same time, in this case it would speed up the searching and
 * indexing process if they are separated in different document bases.
 * <p>
 * A document base uses an indexer to index documents. It uses repositories
 * to store the documents, either indexable ones or attached ones (i.e. non-indexable documents
 * that are logically dependant of the indexable documents, images or the like).
 * An application can get a searcher to perform searches within this document base,
 * possibly with other document bases.
 * <p>
 * In order to work properly, a document base must be instantiated given the following sequence :
 * 1) creation,
 * 2) setting the logger (optional, but suggested for errors messages),
 * 3) configuration,
 * 4) initialization.
 *
 * @see #enableLogging
 * @see #configure
 * @see #init
 *
 */
public class LuceneDocumentBase extends AbstractDocumentBase implements XMLizable {

    /** The list of repositories in this document base. */
    private Hashtable repositories;

    /** The LuceneDatabase for storing in which repository a document exists. */
    //TODO : rename as sdx(Lucene)LookupIndex ? -pb
    private LuceneDatabase sdxRepoIndexDB;

    /** The LuceneDatabase for storing relations between documents. */
    private LuceneDatabase sdxRelationsIndexDB;

    /** A constant for the repository property in the database entities. */
    private final static String REPO_PROPERTY = "repo";

    /** A constant for the repository property in the database entities. */
    private final static String ORIGINAL_DOCUMENT_PROPERTY = "t_doc";

    /** A constant for the path property which contains any attached document ids.*/
    private final static String ATTACHED_DOCS_PROPERTY = "attachedDocs";

    /** A constant for the path property for attached documents. */
    private static final String MIMETYPE_PROPERTY = "mimetype";

    /** The default repository. */
    private Repository defaultRepository = null;

    /**The fields definition object for indexation of documents. */
    private FieldsDefinition fieldsDef;

    /**A lucene index object for searching. */
    private LuceneIndex luceneSearchIndex;

    /**String representation of our repository index directory name. */
    private final String REPO_INDEX_DIRECTORY_NAME = "sdx-repo-index";

    /**String representation of our relations index directory name. */
    private final String RELATIONS_INDEX_DIRECTORY_NAME = "sdx-relations-index";

    /**String representation of our search index directory name. */
    private final String SEARCH_INDEX_DIRECTORY_NAME = "sdx-search-index";

    /**String representation of the "documentBase" attribute name "default". */
    private final String DBELEM_ATTRIBUTE_DEFAULT = "default";

    /**String representation of the element name "fieldList". */
    private final String ELEMENT_NAME_FIELD_LIST = "fieldList";

    /**String representation of the element name "index". */
    private final String ELEMENT_NAME_INDEX = "index";

    /**String representation of the element name "pipeline". */
    private final String ELEMENT_NAME_PIPELINE = "pipeline";

    /**String representation of the element name "luceneSDXInternalFields", declared in sdx.xconf. */
    private final String ELEMENT_NAME_LUCENE_SDX_INTERNAL_FIELDS = "luceneSDXInternalFields";

    /**The document bases id generator*/
    IDGenerator idGen = null;

    /**Prefix for generated id's*/
    private String idPrefix = "";//defaulted

    /**Suffix for generated id's*/
    private String idSuffix = "";//defaulted

    /**Calendar for this indexation times, with default timeZone and locale for now. */
    private GregorianCalendar gCalendar;

    /* String representation for our default pipeline parameter. */
    private final String DOC_URL = "docUrl";

    /* String representation for our pipeline parameter. */
    private final String SDX_USER = "sdxUser";

    /* String representation for our pipeline parameter. */
    private final String SDX_DATE = "sdxDate";

    /* String representation for our pipeline parameter. */
    private final String SDX_FORMATTED_DATE = "sdxFormattedDate";

    /** A constant for the path property for the type of document, Binary, HTML, or XML. */
    private final String DOCUMENT_TYPE = "sdxDocType";

    /**The query parser configuration to use in the init method*/
    private Configuration configuration = null;

    /**Indicates wheter the original document should be stored or not*/
    private boolean keepOriginalDocuments = true; //Defaulted

    /**String representation of the element name "idGenerator".*/
    private final String ELEMENT_NAME_ID_GENERATOR = "idGenerator";

    /**String representation of the attribute name "class".*/
    private final String ATTRIBUTE_CLASS = "class";

    /**String representation of the attribute name "idPrefix".*/
    private final String ATTRIBUTE_ID_PREFIX = "idPrefix";

    /**String representation of the attribute name "idSuffix".*/
    private final String ATTRIBUTE_ID_SUFFIX = "idSuffix";

    /**String representation of the default id generator class name*/
    private final String DEFAULT_ID_GENERATOR_CLASS = "fr.gouv.culture.sdx.documentbase.DefaultIDGenerator";

    /**String representation of the attribute name "keepOriginalDocuments".*/
    private final String DBELEM_ATTRIBUTE_KEEP_ORIGINAL = "keepOriginalDocuments";

    /**String representation of the element name "remote-access". */
    public static final String ATTRIBUTE_NAME_REMOTE_ACCESS = "remote-access";

    /** String representation of our key in the "props" object corresponding a document base's directory location. */
    public static final String DOCUMENTBASE_DIR_PATH = "dbDirPath";

    /**Internal field name for document ids*/
    public static final String INTERNAL_FIELD_NAME_SDXDOCID = "sdxdocid";

    /**Internal field name for sdxall, all document searches*/
    public static final String INTERNAL_FIELD_NAME_SDXALL = "sdxall";

    public static final String INTERNAL_FIELD_VALUE_SDXALL = "1";

    /**Internal field name for document base id*/
    public static final String INTERNAL_FIELD_NAME_SDXDBID = "sdxdbid";

    /**Internal field name for application id*/
    public static final String INTERNAL_FIELD_NAME_SDXAPPID = "sdxappid";

    /**Internal field name for document type*/
    public static final String INTERNAL_FIELD_NAME_SDXDOCTYPE = "sdxdoctype";


    /**Creates the document base.
     *
     * After a document base is created, the logger  could be set (optional, but suggested for errors messages) ;
     * it should then be configured and after, initialized in order to work properly.
     *
     * @see #enableLogging
     * @see #configure
     * @see #init
     */
    public LuceneDocumentBase() {
    }

    /** Sets the configuration options for this document base.
     *
     *
     * @param configuration      The configuration object from which to build a document base.
     * <p>Sample configuration entry:
     <pre>&lt;sdx:documentBase sdx:id = "myDocumentBaseName" sdx:type = "lucene">
     &nbsp;&nbsp;&lt;sdx:fieldList xml:lang = "fr-FR" sdx:variant = "" sdx:analyzerConf = "" sdx:analyzerClass = "">
     &lt;sdx:field code = "fieldName" type = "word" xml:lang = "fr-FR" sdx:analyzerClass = "" sdx:analyzerConf = ""/>
     &lt;sdx:field code = "fieldName2" type = "field" xml:lang = "fr-FR" brief = "true"/>
     &lt;sdx:field code = "fieldName3" type = "date" xml:lang = "fr-FR"/>
     &lt;sdx:field code = "fieldName4" type = "unindexed" xml:lang = "fr-FR"/>
     &lt;/sdx:fieldList>
     &lt;sdx:index>
     &lt;sdx:pipeline sdx:id = "sdxIndexationPipeline">
     &lt;sdx:transformation src = "path to stylesheet, can be absolute or relative to the directory containing this file" sdx:id = "step2" sdx:type = "xslt"/>
     &lt;sdx:transformation src = "path to stylesheet, can be absolute or relative to the directory containing this file" sdx:id = "step3" sdx:type = "xslt" keep = "true"/>
     &lt;/sdx:pipeline>
     &lt;/sdx:index>
     &lt;sdx:repositories>
     &lt;sdx:repository baseDirectory = "blah4" depth = "3" extent = "100" sdx:type = "FS" sdx:default = "true" sdx:id = "blah4"/>
     &lt;sdx:repository ref = "blah2"/>
     &lt;/sdx:repositories>
     &lt;/sdx:documentBase>
     </pre>
     *@see #documented_application.xconf we should link to this in the future when we have better documentation capabilities
     */
    public void configure(Configuration configuration) throws ConfigurationException {
        //verifying the object
        Utilities.checkConfiguration(configuration);
        //retaining the documentbase configuration
        this.configuration = configuration;

        String dbDirPath = "";
        //at this point, we have a <sdx:documentBase> configuration object...
        if (configuration.getName() == DocumentBase.ELEMENT_NAME_DOCUMENT_BASE) {
            try {
                //setting the id from the configuration file
                setId(configuration.getAttribute(DocumentBase.ATTRIBUTE_ID));
            } catch (SDXException sdxE) {
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }
            if (!Utilities.checkString(getId())) {
                String[] args = new String[1];
                args[0] = configuration.getLocation();
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_INVALID_ID_VALUE, args, null);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }
            //getting the path for the location where this document base will reside
            dbDirPath = Utilities.getStringFromHashtable(Application.DOCUMENTBASES_DIR_PATH, props) + getId() + File.separator;
        }
        //... otherwise we have a <sdx:userDocumentBase> element
        else if (configuration.getName() == Application.ELEMENT_NAME_USER_DOCUMENT_BASE) {
            try {
                //setting the id based onto a constant
                setId(Application.USER_DOCUMENT_BASE_ID);
            } catch (SDXException sdxE) {
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }
            //getting the path for the location where this document base will reside
            dbDirPath = Utilities.getStringFromHashtable(Application.USERS_DOCUMENTBASE_DIR_PATH, props);
            //... exit if we have nothing
        } else {
            String[] args = new String[1];
            args[0] = configuration.getName();
            //not passing a logger should be logged later up the stack
            SDXException sdxE = new SDXException(null, SDXExceptionCode.ERROR_UNRECOGNIZED_DOCUMENTBASE_CONFIG, args, null);
            throw new ConfigurationException(sdxE.getMessage(), sdxE);
        }

        //setting property : the path where this document base resides
        props.put(DOCUMENTBASE_DIR_PATH, dbDirPath);

        //setting the default flag, false if not specified in configuration file
        this.isDefault = configuration.getAttributeAsBoolean(DBELEM_ATTRIBUTE_DEFAULT, false);

        this.locale = Utilities.buildLocale(configuration, null);

        this.keepOriginalDocuments = configuration.getAttributeAsBoolean(DBELEM_ATTRIBUTE_KEEP_ORIGINAL, true);

        //at this point, we should have <sdx:fieldList> containing a list of fields
        try {
            //getting the list of fields for indexing
            Configuration fieldList = configuration.getChild(ELEMENT_NAME_FIELD_LIST, false);
            //testing if we have something
            if (fieldList == null) {
                String[] args = new String[1];
                args[0] = configuration.getLocation();
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_NO_FIELD_LIST_CONFIG, args, null);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }

            /*
             AnalyzerManager am = (AnalyzerManager) props.get(FrameworkImpl.ANALYZER_MGR);
             if (am == null) {
                 //if it is null we build a new one, should rarely happen, but just in case
                 am = new AnalyzerManager();
                 //setting the logger for the analyzer manager object
                 am.enableLogging(logger);
                 //passing component manager (actually Cocoon's one) to the analyzer manager object
                 am.compose(manager);
             }
            */

            //assigning a value for the class field
            fieldsDef = new FieldsDefinition();
            //setting the logger
            fieldsDef.enableLogging(this.logger);
            //setting properties
            fieldsDef.setProperties(this.props);
            //configuring the fieldsDefinition with the fields list
            fieldsDef.configure(fieldList);
            //we add SDX internal fields
            Configuration sdxConf = (Configuration) props.get(FrameworkImpl.SDX_CONF);
            //testing if we have something
            if (sdxConf == null) {
                String[] args = new String[2];
                args[0] = FrameworkImpl.CONFIGURATION_FILE_NAME;
                args[1] = Utilities.getStringFromHashtable(FrameworkImpl.SDX_CONF_PATH, props);
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_NO_SDX_CONFIG_FILE, args, null);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }
            Configuration iFConf = sdxConf.getChild(ELEMENT_NAME_LUCENE_SDX_INTERNAL_FIELDS, false);
            //testing if we have something
            if (iFConf == null) {
                //TODOException?:should we be warning here instead of throwing an exception?-rbp
                String[] args = new String[1];
                args[0] = sdxConf.getLocation();
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_NO_INTERNAL_FIELDS_CONFIG, args, null);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }
            //configuring internal field for term searching
            fieldsDef.addInternalFields(iFConf);

        } catch (SDXException e) {
            throw new ConfigurationException(e.getMessage(), e.fillInStackTrace());
        }

        //at this point, we should have a <sdx:pipeline> element
        //getting the pipeline configuration
        //TODO : what if configuration.getChild(ELEMENT_NAME_INDEX) returns null ? -pb*
        //A:it wont't based upon the avalon architecture, please see the avalon javadocs:)-rbp
        //My question was about code design : a code profiler doesn't know of Avalon docs
        //but it does know about potential null pointers -pb
        Configuration pipeConf = configuration.getChild(ELEMENT_NAME_INDEX).getChild(ELEMENT_NAME_PIPELINE, false);
        //testing if we have something
        if (pipeConf == null) {
            //no <sdx:index> or <sdx:pipeline> element
            String[] args = new String[1];
            args[0] = configuration.getLocation();
            SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_NO_PIPELINE_CONFIG, args, null);
            throw new ConfigurationException(sdxE.getMessage(), sdxE);
        }

        //creating the pipeline and assigning the class field
        pipe = new GenericPipeline();
        //setting the logger for the pipeline object
        pipe.enableLogging(this.logger);
        try {
            //passing component manager (actually Cocoon's one) to the pipeline object
            pipe.compose(manager);
            //passing the properties object to the repo
            pipe.setProperties(props);
            //configuring the pipeline object from 'application.xconf'
            pipe.configure(pipeConf);
        } catch (ComponentException e) {
            throw new ConfigurationException(e.getMessage(), e.fillInStackTrace());
        } catch (ConfigurationException e) {
            throw e;
        }

        configureIdGenerator(configuration);

        //at this point, we should have a <sdx:repositories> element containing a list of repositories
        //intializing the array
        //TODO : what if configuration.getChild(Repository.ELEMENT_NAME_REPOSITORIES) returns null ? -pb
        Configuration[] repoConfList = new Configuration[configuration.getChild(Repository.ELEMENT_NAME_REPOSITORIES).getChildren(Repository.ELEMENT_NAME_REPOSITORY).length];
        //getting an array of configuration objects from each of the <sdx:repository> subElements of <sdx:repositories> element
        //TODO : what if configuration.getChild(Repository.ELEMENT_NAME_REPOSITORIES) returns null ? -pb
        //A:it wont't based upon the avalon architecture, please see the avalon javadocs:)-rbp
        //Same answer -pb
        repoConfList = configuration.getChild(Repository.ELEMENT_NAME_REPOSITORIES).getChildren(Repository.ELEMENT_NAME_REPOSITORY);
        //testing if we have something
        if (repoConfList == null || repoConfList.length == 0) {
            String[] args = new String[1];
            //getting the location of the configuration file
            args[0] = configuration.getLocation();
            SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_NO_REPOSITORIES_IN_CONFIG, args, null);
            throw new ConfigurationException(sdxE.getMessage(), sdxE);
        }
        //adding the repositories to this documentbase
        configureRepositories(repoConfList);
    }

    private void configureIdGenerator(Configuration configuration) throws ConfigurationException {
        Configuration idGenConf = configuration.getChild(ELEMENT_NAME_ID_GENERATOR, false);
        if (idGenConf == null)
            this.idGen = new DefaultIDGenerator(); //we use the sdx default
        else {
            //we attempt to use what is given
            String idGenClassName = idGenConf.getAttribute(ATTRIBUTE_CLASS, DEFAULT_ID_GENERATOR_CLASS);
            try {
                Class idGenClass = Class.forName(idGenClassName);
                Object obj = idGenClass.newInstance();
                if (obj == null) {
                    String[] args = new String[1];
                    args[0] = idGenClassName;
                    SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_NEW_OBJECT_INSTANCE_NULL, args, null);
                    throw new ConfigurationException(sdxE.getMessage(), sdxE);
                }
                if (obj instanceof IDGenerator)
                    this.idGen = (IDGenerator) obj;
                else {
                    //the object doesn't implement our interface
                    String[] args = new String[3];
                    args[0] = "IDGenerator";
                    args[1] = idGenClass.getName();
                    args[2] = "IDGenerator";
                    SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_CLASS_NOT_INSTANCEOF_SDX_INTERFACE, null, null);
                    throw new ConfigurationException(sdxE.getMessage(), sdxE);
                }
                this.idGen.compose(this.manager);
            } catch (ClassNotFoundException e) {
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_CONFIGURE_ID_GENERATOR, null, e);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            } catch (InstantiationException e) {
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_CONFIGURE_ID_GENERATOR, null, e);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            } catch (IllegalAccessException e) {
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_CONFIGURE_ID_GENERATOR, null, e);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            } catch (ComponentException e) {
                SDXException sdxE = new SDXException(logger, SDXExceptionCode.ERROR_CONFIGURE_ID_GENERATOR, null, e);
                throw new ConfigurationException(sdxE.getMessage(), sdxE);
            }
            this.idPrefix = idGenConf.getAttribute(ATTRIBUTE_ID_PREFIX, "");
            this.idSuffix = idGenConf.getAttribute(ATTRIBUTE_ID_SUFFIX, "");
        }
    }

    private void configureRepositories(Configuration[] repoConfList) {
        //ensuring we have a hashtable to populate
        //TODO : create the hastable when object is instanciated ? -pb
        if (repositories == null) repositories = new Hashtable();
        //a pointer to the first repository
        Repository firstRepository = null;
        //iterating over the list and creating the repositories
        for (int i = 0; i < repoConfList.length; i++) {
            Repository repo = null;
            try {
                /*check for the ref attribute, if it exists, get the repository object and add it to the local hashtable
                *if the attribute doesn't exist create the repo like below, we also need to handle DEFAULTS with refs*/
                String ref = repoConfList[i].getAttribute(Repository.ATTRIBUTE_REF, null);
                if (Utilities.checkString(ref)) {
                    Hashtable appRepos = (Hashtable) props.get(Application.APPLICATION_REPOSITORIES);
                    if (appRepos != null)
                        repo = (Repository) appRepos.get(ref);
                    if (repo == null) {
                        String[] args = new String[1];
                        args[0] = ref;
                        throw new SDXException(logger, SDXExceptionCode.ERROR_LOAD_REFERENCED_REPO, args, null);
                    }
                    //setting the default flag for the referenced repository
                    repo.setIsDefault(repoConfList[i].getAttributeAsBoolean(Repository.ATTRIBUTE_DEFAULT, false));
                } else
                //creating the repository
                    repo = Utilities.createRepository(repoConfList[i], manager, props, logger);

                //populating the hashtable
                repositories.put(repo.getId(), repo);
                //setting the default repository if applicable
                //TODO : not compliant with SDX 1 behaviour ; first default one out of many was assumed -pb
                //A:it is still the default behavior, if the user does not specify a default repositories-rbp
                if (repo.isDefault()) defaultRepository = repo;
                //retaining first repository
                if (i == 0) firstRepository = repo;
            } catch (SDXException e) {
                //we don't want all repositories configurations to fail so we won't throw this farther out
                //the creation of the SDXException should log this message
            } catch (ConfigurationException e) {
                //we don't want all repository configurations to fail so we won't throw this farther out
                Utilities.logException(logger, e);
            }
        }

        //if no default repository was provided, we set the default to the first repository
        if (defaultRepository == null) defaultRepository = firstRepository;

    }

    /** Gets a repository in this document base.
     *
     * @param id The repository's id, if <code>null</code> for the default repository is returned
     * @return The repository object
     */
    public Repository getRepository(String id) throws SDXException {
        if (id == null)
            return defaultRepository;
        else {
            Repository repo = (Repository) repositories.get(id);
            if (repo == null) {
                String[] args = new String[2];
                args[0] = id;
                args[1] = this.id;
                throw new SDXException(logger, SDXExceptionCode.ERROR_UNKNOWN_REPOSITORY, args, null);
            } else
                return repo;
        }
    }

    /** Returns the default repository for this document base.
     * @return The default repository object.
     */
    public Repository getDefaultRepository() {
        return defaultRepository;
    }

    /**
     * Feeds a SAX content handler with information about this document base.
     *
     * NOT IMPLEMENTED YET!
     *	@param	hdl		A SAX content handler to feed with events.
     */
    public void toSAX(ContentHandler hdl) {
        // TODO: what will be the XML structure?
        //Export private variables ? Call their toSAX() method when applicable ? -pb
    }

    /** Adds a document.
     *
     * @param doc   The document to add.
     * @param repository    The repository where to store the document.
     * @param params	The parameters for this adding action.
     * @param handler	A content handler where to send information about the process (may be null), currently we don't use it, but maybe in the future.
     * TODO : what kind of "informations" ? -pb
     */
    public synchronized void index(IndexableDocument doc, Repository repository, IndexParameters params, ContentHandler handler) throws SDXException, SAXException, ProcessingException {
        IndexableDocument[] docs = new IndexableDocument[1];
        docs[0] = doc;
        index(docs, repository, params, handler);
        //TODO : if connections are pooled, release them here -pb
    }

    /** Adds a document from an array of documents using the same connection.
     *
     * @param doc			The document to add.
     * @param repository	The repository where to store the document.
     * @param conn          The repository's connection, we want to use only one connection for all the documents
     * @param params     The parameters for this adding action.
     * @param handler	A content handler where to send information about the process (may be null), currently we don't use it, but maybe in the future.
     * TODO : what kind of "informations" ? -pb
     */
    private synchronized void index(IndexableDocument doc, Repository repository, RepositoryConnection conn, IndexParameters params, ContentHandler handler, boolean batchIndex) throws SDXException, SAXException, ProcessingException {
        //ensuring we have valid objects
        super.index(doc, repository, params, handler);

        //we always use our document base's logger during indexation, if another has been provided it will not be used
        doc.enableLogging(this.logger);

        //setting the id generator properties
        doc.setIdGenerator(this.idGen, this.idPrefix, this.idSuffix);

        /*
        * setting the document's url as a pipe parameter
        * we should always have pipelineParams at this point
        * because setBaseParameters has already been called
        */
        java.net.URL tmpURL = doc.getURL(); //Quick patch to increase performance -pb
        Parameters tmpParameters = params.getPipelineParams(); //Quick patch to increase performance -pb
        if (tmpURL != null)
            tmpParameters.setParameter(DOC_URL, tmpURL.toExternalForm());
        else
            tmpParameters.setParameter(DOC_URL, "");

        //setting the parameters to the pipeline
        //TODO?:what if this has already been done outside of our code, we reset them after adding our infos,
        //I think it is ok for now, but maybe bad in some complex cases-rbp
        fr.gouv.culture.sdx.pipeline.Pipeline tmpPipeline = params.getPipeline();  //Quick patch to increase performance -pb        
        tmpPipeline.setParameters(tmpParameters);

        //indexing document using the pipeline from the index params
        tmpPipeline.setConsumer(doc);
        Parser parser = null;
        try {
            parser = (Parser) this.manager.lookup(Parser.ROLE);
            doc.startIndexing(parser, tmpPipeline);
            //after parsing we want to retrieve any transformed documents
            byte[] tmpBytes = tmpPipeline.getTransformedBytes(); //Quick patch to increase performance -pb 
            java.io.File tmpFile = tmpPipeline.getTransformedFile(); //Quick patch to increase performance -pb
            if (tmpBytes != null)
                doc.setTransformedDocument(tmpBytes);
            else if (tmpFile != null)
                doc.setTransformedDocument(tmpFile);
            //TODO : potential other case here -pb
        } catch (ComponentException e) {
            String[] args = new String[1];
            args[0] = e.getMessage();
            throw new SDXException(logger, SDXExceptionCode.ERROR_ACQUIRE_PARSER, args, e);
        } finally {
            if (parser != null) this.manager.release(parser);
        }

        //checking to see if the doc exists in the repositories via the repository look-up index
        DatabaseEntity lookupEntry = sdxRepoIndexDB.getEntity(doc.getId());
       
        String status = "";

        //doc doesn't exist... adding
        if (lookupEntry == null) {
            //TODO : instantiate the connection here or get it from a pool at a global level -pb
            add(doc, repository, conn, params, batchIndex, true);            
        }
        //doc exists... handling using params
        else {
            lookupEntry.enableLogging(this.logger);
            switch (params.handleSameId()) {
                case IndexParameters.SAME_ID_ERROR:
                    String repoId = lookupEntry.getProperty(REPO_PROPERTY);
                    if (logger != null) logger.error("this document already exists within the repository : " + repoId + " of your documentbase");
                    //TODO:generate the proper exception here and send the toSAX on the exception.
                    //addAtts.addAttribute("", "status", "status", "CDATA", "exists");
                    //TODOLogging: send something to the handler in the future, but do not throw an exception
                    status = "failure";
                    break;
                case IndexParameters.SAME_ID_IGNORE:
                    //TODOLogging: see with Martin -- where to put this string?
                    Utilities.logInfo(logger, "ignoring document : " + doc.getId());
                    status = "ignored";
                    break;
                case IndexParameters.SAME_ID_REPLACE:
                default:
                    //TODO : instantiate the connection here or get it from a pool at a global level -pb
                    delete(doc, batchIndex, handler, true);
                    //then adding it according to its new properties
                    add(doc, repository, conn, params, batchIndex, true);
                    status = "replacement";
                    break;
            }
        } 
        //TODO : if connections are pooled, release them here -pb

        if (handler != null && params.getSendIndexationEvents() == IndexParameters.SEND_ALL_EVENTS) {
            AttributesImpl addAtts = new AttributesImpl();
            addAtts.addAttribute("", Node.Name.STATUS, Node.Name.STATUS, "CDATA", status);
            addAtts.addAttribute("", Node.Name.ID, Node.Name.ID, "CDATA", doc.getId());
            addAtts.addAttribute("", Node.Name.REPO, Node.Name.REPO, "CDATA", repository.getId());
            addAtts.addAttribute("", Node.Name.BASE, Node.Name.BASE, "CDATA", this.getId());
            String appId = Utilities.getStringFromHashtable(Application.APPLICATION_ID, props);
            if (Utilities.checkString(appId))
                addAtts.addAttribute("", Node.Name.APP, Node.Name.APP, "CDATA", appId);
            handler.startElement(Framework.SDXNamespaceURI, Node.Name.UPLOAD_DOCUMENT, Framework.SDXNamespacePrefix + ":" + Node.Name.UPLOAD_DOCUMENT, addAtts);
            handler.endElement(Framework.SDXNamespaceURI, Node.Name.UPLOAD_DOCUMENT, Framework.SDXNamespacePrefix + ":" + Node.Name.UPLOAD_DOCUMENT);
        }        
    }

    /** Adds some documents.
     *
     * @param docs The documents to add.
     * @param repository The repository where to store the documents. If null is passed, the default repository will be used.
     * @param params The parameters for this adding action.
     * @param handler	A content handler where to send information about the process (may be null)
     * TODO : what kind of "informations" ? -pb
     */
    public synchronized void index(IndexableDocument[] docs, Repository repository, IndexParameters params, ContentHandler handler) throws SDXException, SAXException, ProcessingException {

        /** boolean indicating a batch indexation so the the search
         * index can be optimized only once at the end of the batch,
         * false by default to indicate single document indexation.
         */
        //ensuring we have valid parameters
        params = setBaseParameters(params);

        if (handler != null) handler.startElement(Framework.SDXNamespaceURI, Node.Name.UPLOAD_DOCUMENTS, Framework.SDXNamespacePrefix + ":" + Node.Name.UPLOAD_DOCUMENTS, new AttributesImpl());

        long start = System.currentTimeMillis();
        boolean batchIndex = false;
        int docsAdded = 0;
        int docsFailed = 0;

        //if we have a very big patch we split them into smaller batches
        if (docs != null && docs.length > 0) {
            double batchMax = params.getBatchMax();
            //we reset the batch max if it is a smaller batch
            if (batchMax > docs.length) batchMax = docs.length;
            //we calculate the number of batches needed
            int batches = (int) Math.ceil((docs.length / batchMax));
            //setting the index of the document for indexing in the docs array
            int docIdx = 0;
            //selecting a batch
            for (int i = 0; i < batches; i++) {
                //we have batch of documents to index so we set a true state
                batchIndex = true;
                //using the same connection to the store for each document
                RepositoryConnection conn = null;
                //if no repository is provided, using the default repository
                if (repository == null) {
                    String[] args = new String[1];
                    args[0] = this.id;
                    SDXException sdxE = new SDXException(null, SDXExceptionCode.ERROR_USING_DEFAULT_REPO, args, null);
                    Utilities.logInfo(logger, sdxE.getMessage());
                    //we are using the default because no repository was found
                    repository = defaultRepository;
                }

                //ensuring we have a valid repository id to work with
                if (!Utilities.checkString(repository.getId()))
                    throw new SDXException(logger, SDXExceptionCode.ERROR_INVALID_REPO_ID, null, null);

                //setting our base params
                params = setBaseParameters(params);

                try {

                    conn = repository.getConnection();

                    //indexing documents upto batchMax with the same connection
                    for (int k = 1; docIdx < docs.length && k <= batchMax; k++) {
                        //if we have the last document we set the flag that the batch index and delete are finished, so we can optimize
                        if ((k % batchMax) == 0 || (docIdx == docs.length - 1))
                            batchIndex = false;

                        try {
                            //indexing the next document
                            index(docs[docIdx], repository, conn, params, handler, batchIndex);
                            docsAdded++;
                        } catch (SDXException e) {
                            docsFailed++;
                            if (handler != null && params.getSendIndexationEvents() >= IndexParameters.SEND_ERRORS_EVENTS) {
                                AttributesImpl failAtts = new AttributesImpl();
                                failAtts.addAttribute("", Node.Name.STATUS, Node.Name.STATUS, "CDATA", "failure");
                                if (docs[docIdx] != null)
                                    failAtts.addAttribute("", Node.Name.ID, Node.Name.ID, "CDATA", docs[docIdx].getId());
                                if (repository != null)
                                    failAtts.addAttribute("", Node.Name.REPO, Node.Name.REPO, "CDATA", repository.getId());
                                failAtts.addAttribute("", Node.Name.BASE, Node.Name.BASE, "CDATA", this.getId());
                                String appId = Utilities.getStringFromHashtable(Application.APPLICATION_ID, props);
                                if (Utilities.checkString(appId))
                                    failAtts.addAttribute("", Node.Name.APP, Node.Name.APP, "CDATA", appId);
                                handler.startElement(Framework.SDXNamespaceURI, Node.Name.UPLOAD_DOCUMENT, Framework.SDXNamespacePrefix + ":" + Node.Name.UPLOAD_DOCUMENT, failAtts);
                                e.toSAX(handler);
                                handler.endElement(Framework.SDXNamespaceURI, Node.Name.UPLOAD_DOCUMENT, Framework.SDXNamespacePrefix + ":" + Node.Name.UPLOAD_DOCUMENT);
                            }
                            //catch and then continue adding other docs, message should be already logged, but pass some info's to the handler in the future
                        } finally {
                            //incrementing to the next document
                            docIdx++;
                        }

                    }
                    //TODOLogging:add some summary info and info for each document uploaded successfully

                } catch (SDXException e) {
                    //catch and then continue adding other docs, message should be already logged, but pass some info's to the handler in the future
                } finally {
                    //conn.optimize();
                    //releasing the connection
                    repository.releaseConnection(conn);
                    compactSearchIndex();                    
                }
            }
        }
        long finish = System.currentTimeMillis();
        long elapsed = finish - start;

        if (handler != null && params.getSendIndexationEvents() >= IndexParameters.SEND_STATS_EVENTS) {
            AttributesImpl summaryAtts = new AttributesImpl();
            summaryAtts.addAttribute("", Node.Name.ADDITIONS, Node.Name.ADDITIONS, "CDATA", Integer.toString(docsAdded));
            // summaryAtts.addAttribute("", "replacements", "replacements", "CDATA", Integer.toString(docsAdded));
            summaryAtts.addAttribute("", Node.Name.FAILURES, Node.Name.FAILURES, "CDATA", Integer.toString(docsFailed));
            summaryAtts.addAttribute("", Node.Name.DURATION, Node.Name.DURATION, "CDATA", Long.toString(elapsed / 1000));
            handler.startElement(Framework.SDXNamespaceURI, "summary", Framework.SDXNamespacePrefix + ":summary", summaryAtts);
            handler.endElement(Framework.SDXNamespaceURI, "summary", Framework.SDXNamespacePrefix + ":summary");
        }

        if (handler != null) handler.endElement(Framework.SDXNamespaceURI, Node.Name.UPLOAD_DOCUMENTS, Framework.SDXNamespacePrefix + ":" + Node.Name.UPLOAD_DOCUMENTS);

    }

    //TODO : possibly move this method to the LuceneIndex class -pb
    private void compactSearchIndex() throws SDXException {        
        luceneSearchIndex.mergeBatch(); //TODO : make sure there is no sync problems -pb
    }

    /**Set's the default pipeline parameters and ensures the params have a pipeline
     *
     * @param params The params object provided by the user at indexation time
     */
    private IndexParameters setBaseParameters(IndexParameters params) {
        //if we get null params, we create default ones
        if (params == null)
            params = new IndexParameters();
        //otherwise we check if we have lucene params and set them to the index
        else if (params instanceof LuceneIndexParameters)
            setSearchIndexParameters((LuceneIndexParameters) params);

        //providing a logger for the params
        params.enableLogging(logger);

        //checking pipeline parameters,
        if (params.getPipelineParams() == null)
            params.setPipelineParams(new Parameters());

        //adding the user, if it exists
        fr.gouv.culture.sdx.user.User tmpUser = params.getUser(); //Quick patch to increase performance -pb
        if (tmpUser != null)
            params.getPipelineParams().setParameter(SDX_USER, tmpUser.getId());
        //adding our defaults
        //getting current date info
        Date date = gCalendar.getTime();
        //formmating a humanReadable date
        params.getPipelineParams().setParameter(SDX_DATE, DateFormat.getDateTimeInstance(DateFormat.FULL, DateFormat.FULL).format(date));
        //formatting an iso 8601 date
        //TODORefactor: this date format into our date utilities-rbp
        // params.getPipelineParams().setParameter(SDX_FORMATTED_DATE, new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssz").format(date));
        // FG: to get exact system date, formatted by our date utilities
        params.getPipelineParams().setParameter(SDX_FORMATTED_DATE, fr.gouv.culture.sdx.utils.Date.formatDate(new Date()));
        //if their is not a pipline specified we use our default pipeline from the application configuration
        //TODO?:what if this has already been done outside of our code, we reset them after adding our infos, it think it is ok?-rbp
        /*raises problems when the developer reUses the same index parameters for different document bases, once set after the below code,
        then we have no handle on its validity as the correct pipeline*/
        if (params.getPipeline() == null) params.setPipeline(this.pipe);

        return params;
    }
    
    //TODO : possibly move this method to the LuceneIndex class -pb
    private void setSearchIndexParameters(LuceneIndexParameters params) {
        luceneSearchIndex.setParameters(params);
    }

    /**
     * Deletes a document and any attached document(s) if not used by any other document(s).
     *
     *@param	doc The document to delete.
     *      * @param   handler A content handler to feed with information.

     * @throws SDXException
     */
    public synchronized void delete(Document doc, ContentHandler handler) throws SDXException, SAXException, ProcessingException {
        //passing false to indicate we don't have batch to delete
        Document[] docs = new Document[1];
        docs[0] = doc;
        delete(docs, handler);
    }

    /**
     * Deletes an array of documents and any attached document(s) if not used by any other document(s).
     *
     * @param	docs	The documents to delete.
     * @param   handler A content handler to feed with information.
     * @throws SDXException
     *
     */
    public synchronized void delete(Document[] docs, ContentHandler handler) throws SDXException, SAXException, ProcessingException {
        /** boolean indicating a batch deletion so the the search
         * index can be optimized only once at the end of the batch,
         * false by default to indicate single document deletion.
         */
        boolean batchDelete;

        /* TODORefactor:
        in the future we should implement some logic to lookup the documents and group them by repository,
        get a connection for each repo and then delete using the same connection
         */
        try {
            if (docs != null) {
                //we a batch of documents to delete so we set a true state
                batchDelete = true;
                if (handler != null) {
                    AttributesImpl atts = new AttributesImpl();
                    String appId = Utilities.getStringFromHashtable(Application.APPLICATION_ID, props);
                    if (Utilities.checkString(appId))
                        atts.addAttribute("", Node.Name.APP, Node.Name.APP, "CDATA", appId);
                    atts.addAttribute("", Node.Name.BASE, Node.Name.DB_ID, "CDATA", this.getId());
                    handler.startElement(Framework.SDXNamespaceURI, Node.Name.DELETIONS, Framework.SDXNamespacePrefix + ":" + Node.Name.DELETIONS, atts);
                }
                for (int i = 0; i < docs.length; i++) {
                    //if we have the last document we set the flag that the batch delete is finished
                    if (i == (docs.length - 1)) batchDelete = false;
                    delete(docs[i], batchDelete, handler, true);
                    if (handler != null) {
                        AttributesImpl deleteAtts = new AttributesImpl();
                        deleteAtts.addAttribute("", Node.Name.NO, Node.Name.NO, "CDATA", Integer.toString(i + 1));
                        deleteAtts.addAttribute("", Node.Name.ID, Node.Name.ID, "CDATA", docs[i].getId());
                        handler.startElement(Framework.SDXNamespaceURI, Node.Name.DELETION, Framework.SDXNamespacePrefix + ":" + Node.Name.DELETION, deleteAtts);
                        handler.endElement(Framework.SDXNamespaceURI, Node.Name.DELETION, Framework.SDXNamespacePrefix + ":" + Node.Name.DELETION);
                    }
                }
            }
        } catch (SDXException e) {
            if (handler != null)
                e.toSAX(handler);
            else
                throw e;
        } finally {
            if (handler != null)
                handler.endElement(Framework.SDXNamespaceURI, Node.Name.DELETIONS, Framework.SDXNamespacePrefix + ":" + Node.Name.DELETIONS);
        }
    }

    /**
     * Deletes a document and any attached document(s) if not used by any other document(s).
     *
     *@param	doc The document to delete.
     * @throws SDXException
     */
    private synchronized void delete(Document doc, boolean batchDelete, ContentHandler handler, boolean isMaster) throws SDXException, SAXException, ProcessingException {
        //ensuring we have valid objects
        Utilities.checkDocument(logger, doc);
        Repository repository = null;
        RepositoryConnection conn = null;
        try {            

            //TODO : refactor connection management (in the calling method ? use a pool ?)                  
            //some set-up for accessing the repository
            repository = getRepositoryForDocument(doc);
            conn = repository.getConnection();         

            //ensuring we have valid objects
            super.delete(doc, null);

            /*
            * this flag replaces the tests on (doc instanceof IndexableDocument)
            * some instances of IndexableDocument (e.g. TransformedDocuments) are *not* to be indexed)            * 
            * TODO : find a better name and improve concept -pb
            */
            if (isMaster) { 

                //casting the document to an indexable document
                IndexableDocument indexableDoc = (IndexableDocument)doc;

                /*
                * STEP 1 : deleteing the document's indexation from the search index
                */                
                //TODO : test if the deletion is relevant since the document may not be indexed *yet* -pb
                deleteFromSearchIndex(indexableDoc.getId(), batchDelete);

                //some set-up for document's dependencies
                DatabaseEntity docLookupEntry = sdxRepoIndexDB.getEntity(indexableDoc.getId());
                //TODO : check docLookupEntry validity : it normaly can't be null -pb
                docLookupEntry.enableLogging(logger);
                /*
                * TODO : provide a getOriginalDocument() method to AbstractIndexableDocument ? 
                * this could avoid to instantiate a DatabaseEntity for retrieving the original document -pb
                */
                //Warning : given the behaviour of DatabaseEntity.getProperty, only 1 original document is assumed -pb
                String origDocId = docLookupEntry.getProperty(ORIGINAL_DOCUMENT_PROPERTY);

                /*
                * STEP 2 : deleteing the original document, if relevant
                */                
                if (origDocId != null) {
                    Document origDoc = new BinaryDocument(); //Sic ! -pb
                    origDoc.enableLogging(logger);
                    origDoc.setId(origDocId);                    
                    //Notice that the original document look-up entry will be deleted as well                    
                    this.delete(origDoc, batchDelete, handler, false); //recursive call

                    //TODO : uncomment when approved -pb
                    //deleteing the relation between the document and the original document
                    //deleteFromRelationsIndex(doc, origDoc, new String("original"), 1, batchDelete);
                }

                /*
                * STEP 3 : deleteing the attached documents, if relevant
                */   
                String[] attachedDocIds = docLookupEntry.getPropertyValues(ATTACHED_DOCS_PROPERTY);
                for (int i = 0; attachedDocIds != null && i < attachedDocIds.length; i++) {
                    //getting the attached docucment
                    BinaryDocument attachedDoc = new BinaryDocument(); 
                    attachedDoc.enableLogging(logger);
                    attachedDoc.setId(attachedDocIds[i]);
                    DatabaseEntity owners[] = getOwners(attachedDocIds[i]); 
                    //Delete attached document entry if it is owned by our doc and our doc only
                    if (owners == null) {
                        //TODOException : the attached document is not owned !
                    }
                    else if (owners.length == 1 && owners[0].getId().equals(indexableDoc.getId())) {                                
                       //Notice that the attached document look-up entry will be deleted as well
                       this.delete(attachedDoc, batchDelete, handler, false); //recursive call 
                    }
                    else if (owners.length == 1 && !owners[0].getId().equals(indexableDoc.getId())) {
                        //TODOException : the attached document is not owned by our document !
                    }
                    else
                    {                                                                        
                        //Attached document is owned by more than 1 document : keep it
                        //TODO : check that our document is among the oners
                    }
                    //TODO : uncomment when approved -pb
                    //deleteing the relation between the document and the attached document
                    //deleteFromRelationsIndex(doc, attachedDoc, new String("attached"), 1, batchDelete);
                }
            }

            /*
            * STEP 4 : delete the document's entry from the ook-up index
            */    
            deleteFromLookupIndex(doc, batchDelete);

            /*
            * STEP 5 : deleteing the document itself
            */    
            repository.delete(doc, conn);

            /*
            * TODO : that should never be the case but it would be safe to destroy all other relations
            * to and from this document, using for example :
            * deleteFromRelationsIndex(doc, null, null, 0, batchDelete);
            */
            

        } catch (SDXException e) {
            throw e;
        } finally {
            //always close the connection
            if (repository != null && conn != null) repository.releaseConnection(conn);
            //TODO : delegate the release to the calling method -pb
        }
    }
    
    //TODO : possibly move this method to a LuceneLookupIndex class or to LuceneDatase -pb
    private void deleteFromLookupIndex(Document doc, boolean batchDelete) throws SDXException {
        try {
            DatabaseEntity docLookupEntry = sdxRepoIndexDB.getEntity(doc.getId());
            //TODO : check docLookupEntry validity -pb
            docLookupEntry.enableLogging(logger);
            sdxRepoIndexDB.delete(docLookupEntry);
            //TODO : transfer a batch control here -pb
        } catch (SDXException e) {
            throw e;
        }    
    }

    //TODO : possibly move this method to a LuceneLookupIndex class or to LuceneDatase -pb
    private void deleteFromRelationsIndex(Document docA, Document docB, String relationType, int relationDirection, boolean batchDelete) throws SDXException {
        try {            
            TermQuery tq1 = new TermQuery(new Term("docAID", docA.getId()));
            //TODO : consider (docB == null) when direction is "both" : it would provide a powerful mechanism to destroy *all* relations -pb
            //TODO : add repository when we have a Document.getRepository() method -pb
            TermQuery tq3 = new TermQuery(new Term("docBID", docB.getId()));
            //TODO : add repository when we have a Document.getRepository() method -pb
            //TODO : consider (relationType == null) that would mean "all types" -pb
            TermQuery tq5 = new TermQuery(new Term("type", relationType));            
            //Relation order is not significant
            TermQuery tq6 = new TermQuery(new Term("direction", String.valueOf(relationDirection)));                        
            //Make a first query
            org.apache.lucene.search.BooleanQuery bq1 = new org.apache.lucene.search.BooleanQuery();
            //all terms are mandatory
            bq1.add(tq1, true, false);
            bq1.add(tq3, true, false);
            bq1.add(tq5, true, false);
            bq1.add(tq6, true, false);
            TermQuery tq7 = new TermQuery(new Term("docAID", docB.getId()));
            //TODO : add repository when we have a Document.getRepository() method -pb
            TermQuery tq9 = new TermQuery(new Term("docBID", docA.getId()));
            //TODO : add repository when we have a Document.getRepository() method -pb
            //TODO : consider (relationType == null) that would mean "all types" -pb
            TermQuery tq11 = new TermQuery(new Term("type", relationType));            
            //Relation order is not significant
            TermQuery tq12 = null;
            switch (relationDirection) {
                case 0: //bidirectionnal ; TODO : use constants
                    tq12 = new TermQuery(new Term("direction", String.valueOf(0)));
                case 1 : //A to B ; TODO : use constants
                    tq12 = new TermQuery(new Term("direction", String.valueOf(2)));
                case 2 : //A from B ; TODO : use constants 
                    tq12 = new TermQuery(new Term("direction", String.valueOf(1)));
                default :
                    //TODO : throw exception
            }
            //Make a second query, this time with inverted A and B documents
            org.apache.lucene.search.BooleanQuery bq2 = new org.apache.lucene.search.BooleanQuery();
            //all terms are mandatory
            bq2.add(tq7, true, false);
            bq2.add(tq9, true, false);
            bq2.add(tq11, true, false);
            bq2.add(tq12, true, false);
            //Now, do a logical OR for the 2 queries
            org.apache.lucene.search.BooleanQuery bq = new org.apache.lucene.search.BooleanQuery();
            bq.add(bq1, false, false);
            bq.add(bq2, false, false);
            //retrieving a list of docs which use this same attached documents
            Hits h = sdxRelationsIndexDB.search(bq);
            if (h == null) 
                {} //TODO : determinate the required behaviour : exception, do nothing...
            else
            {            
                if (h.length() != 1)
                {
                    //TODO : exception, more than 1 relation of the same type between A and B should be impossible
                }
                try {
                    DatabaseEntity relationEntry = sdxRelationsIndexDB.getEntity(h.doc(1));
                    //TODO : check relationEntry validity -pb
                    relationEntry.enableLogging(logger);
                    sdxRepoIndexDB.delete(relationEntry);

                    logger.info("deleted relation : " + relationEntry.getId());

                    //TODO : should we update the subsequent relationPosition ? -pb
                    //TODO : transfer a batch control here -pb

                } catch (IOException e) {
                    //could not retrieve the document
                    String[] args = new String[2];
                    args[0] = sdxRepoIndexDB.getIndexPath();
                    args[1] = e.getMessage();
                    throw new SDXException(logger, SDXExceptionCode.ERROR_LUCENE_RETRIEVE_DOCUMENT, args, e);
                }
           }
        } catch (SDXException e) {
            throw e;
        }    
    }

    //TODO : possibly move this method to the LuceneIndex class -pb
    private void deleteFromSearchIndex(String docId, boolean batchDelete) throws SDXException {
        try {
            luceneSearchIndex.deleteDocument(docId, batchDelete);
            //TODO : transfer a batch control here -pb
        } catch (SDXException e) {
            throw e;
        }    
    }
    
    //TODO : possibly move this method to a LuceneLookupIndex class or to LuceneDatase -pb
    //TODO : return a String[] with documents' Ids rather than entities ? -pb
    private DatabaseEntity[] getOwners(String attachedDocId) throws SDXException {
        try {
            Term t = new Term(ATTACHED_DOCS_PROPERTY, attachedDocId);
            //creating a query for the term
            TermQuery tq = new TermQuery(t); 
            //retrieving a list of docs which use this same attached documents
            Hits h = sdxRepoIndexDB.search(tq);
            if (h == null) 
                return null; //Could also have returned new DatabaseEntity[0]
            else
            {            
                DatabaseEntity owners[] = new DatabaseEntity[h.length()];
                //String owners[] = new String[h.length()];

                for (int i = 0; i < h.length(); i++) {
                    DatabaseEntity lookupEntry = null;                
                    try {
                        lookupEntry = sdxRepoIndexDB.getEntity(h.doc(i));
                        lookupEntry.enableLogging(logger);
                        owners[i] = lookupEntry;
                        //owners[i] = lookupEntry.getId();
                    } catch (IOException e) {
                        //could not retrieve the document for the int "i" from the hits for attached documents
                        String[] args = new String[2];
                        args[0] = sdxRepoIndexDB.getIndexPath();
                        args[1] = e.getMessage();
                        throw new SDXException(logger, SDXExceptionCode.ERROR_LUCENE_RETRIEVE_DOCUMENT, args, e);
                    }
                }
                return owners;
            }
        } catch (SDXException e) {
            throw e;
        }    
    }

    /*
     *	Deletes the documents found with a query.
     *
     *	@param	res		The found documents to delete.
     */
    //  public void delete(SDXResults res) {
    /* TODO

	    Same as above, we obtain the documents from the results, with a loop using the same store connection.

	 */
    //  }


    /** Initializes the document base and makes the necessary data structures available.
     *
     * The method creates an internal gregorian calendar, an directory for the document base data,
     * directorires for the repository index and search index,
     * @throws SDXException
     */
    public void init() throws SDXException, ConfigurationException {

        //creating and configuring the calendar
        //TODOLocale:we will want to provide the locale from our configuration in the future-rbp
        gCalendar = new GregorianCalendar();

        //TODOPerfomance : performance issue, uncomment the below variable and use it for building directories objects ? -pb
        //String dbDirPath = props.getProperty(DOCUMENTBASE_DIR_PATH);

        //creating an upper level directory for this document base
        //by using our document base dir path from the properties object, as this way we are sure it is the same each time
        //testing the directory, to ensure it is available and we have access
        Utilities.checkDirectory(Utilities.getStringFromHashtable(DOCUMENTBASE_DIR_PATH, props), logger);

        //creating the repository index directory
        //by using our document base dir path from the properties object, as this way we are sure it is the same each time
        String sdxRepoIndexDirPath = Utilities.getStringFromHashtable(DOCUMENTBASE_DIR_PATH, props) + REPO_INDEX_DIRECTORY_NAME + File.separator;
        //testing the directory, to ensure it is available and we have access to it
        File sdxRepoIndexDir = Utilities.checkDirectory(sdxRepoIndexDirPath, logger);

        //initializing the repository look-up index
        //TODO : possibly move this section to a LuceneLookupIndex class or to LuceneDatase -pb
        sdxRepoIndexDB = new LuceneDatabase(sdxRepoIndexDir);
        sdxRepoIndexDB.enableLogging(this.logger);
        sdxRepoIndexDB.init();

        //creating the relations index directory
        //by using our document base dir path from the properties object, as this way we are sure it is the same each time
        String sdxRelationsIndexDirPath = Utilities.getStringFromHashtable(DOCUMENTBASE_DIR_PATH, props) + RELATIONS_INDEX_DIRECTORY_NAME + File.separator;
        //testing the directory, to ensure it is available and we have access to it
        File sdxRelationsIndexDir = Utilities.checkDirectory(sdxRelationsIndexDirPath, logger);

        //initializing the repository look-up index
        //TODO : possibly move this section to a LuceneLookupIndex class or to LuceneDatase -pb
        sdxRelationsIndexDB = new LuceneDatabase(sdxRelationsIndexDir);
        sdxRelationsIndexDB.enableLogging(this.logger);
        sdxRelationsIndexDB.init();

        //creating the search index directory
        //by using our document base dir path from the properties object, as this way we are sure it is the same each time
        String sdxSearchIndexPath = Utilities.getStringFromHashtable(DOCUMENTBASE_DIR_PATH, props) + SEARCH_INDEX_DIRECTORY_NAME + File.separator;
        //testing the directory, to ensure it is available and we have access
        File sdxSearchIndexDir = Utilities.checkDirectory(sdxSearchIndexPath, logger);

        //initializing the search index
        String rmiHost = Utilities.getStringFromHashtable(FrameworkImpl.RMI_HOST, props);
        Integer rmiPort = (Integer) props.get(FrameworkImpl.RMI_PORT);
        String appId = Utilities.getStringFromHashtable(Application.APPLICATION_ID, props);
        initSearchIndex(sdxSearchIndexDir, rmiHost, rmiPort, appId, this.getId());
    }

    //TODO : possibly move this method to the LuceneIndex class -pb
    private void initSearchIndex(File searchIndexDir, String rmiHost, Integer rmiPort, String appId, String documentbaseId) throws SDXException,ConfigurationException {
        //building the search index
        luceneSearchIndex = new LuceneIndex(searchIndexDir, rmiHost, rmiPort, appId, documentbaseId);
        luceneSearchIndex.enableLogging(this.logger);
        luceneSearchIndex.configure(this.configuration);
        //creating a metaAnalyzer for the search index
        MetaAnalyzer mAnalyzer = new MetaAnalyzer();
        //setting the logger
        mAnalyzer.enableLogging(logger);
        //setting up the metaAnalyzer
        mAnalyzer.setUp(fieldsDef);
        //passing the metaAnalyzer to the search index
        luceneSearchIndex.setMetaAnalyzer(mAnalyzer);
        //initializing search index
        luceneSearchIndex.init();
    }

    /** Gets the LuceneIndex object for indexing and searching.
     * @return The LuceneIndex object.
     */
    public LuceneIndex getLuceneIndex() {
        return luceneSearchIndex; //TODO : this method seems to be only used by sdx.xsl -pb
    }

    /** Gets a SDX document as SAX events.
     *
     * @param doc A ParsableDocument, ie XMLDocument or HTMLDocument.
     * @param consumer A SAX content handler to feed with events.
     * <p>
     * The wrapped contentHandler for including events within an XSP page contentHandler should be created using
     * <CODE>IncludeXMLConsumer stripper = new IncludeXMLConsumer(xspContentHandler);</CODE>
     * @throws SDXException
     */
    public void getDocument(ParsableDocument doc, XMLConsumer consumer) throws SDXException {
        //ensuring we have valid objects
        super.getDocument(doc, consumer);
        Repository repo = null;
        RepositoryConnection conn = null;

        try {
            //getting the repository in which the document resides
            repo = getRepositoryForDocument(doc);
            //getting a connection for the repository
            conn = repo.getConnection();
            //then call toSAX from the repository
            repo.toSAX(doc, consumer, conn);
        } catch (SDXException e) {
            throw e;
        } finally {
            //always close the connection
            if (repo != null && conn != null) repo.releaseConnection(conn);
        }
    }

    /** Gets a SDX document as SAX events.
     *
     * @param doc           A Document, ie XMLDocument or HTMLDocument, we discern the type from the document lookup.
     * @param consumer      A SAX content handler to feed with events.
     * @param docTypeKnown   If the type of the ParsableDocument desired is unknown, ie. XML or HTML,
     *                      allows users to build one or the other type of document and still retrieve a document
     * <p>
     * The wrapped contentHandler for including events within an XSP page contentHandler should be created using
     * <CODE>IncludeXMLConsumer stripper = new IncludeXMLConsumer(xspContentHandler);</CODE>
     * @throws SDXException
     */
    //added this one for Frederic
    //TODO?:would this be valid for the DocumentBase interface?
    public void getDocument(ParsableDocument doc, XMLConsumer consumer, boolean docTypeKnown) throws SDXException {
        //ensuring we have valid objects
        super.getDocument(doc, consumer);
        if (!docTypeKnown) {
            //looking-up for the document in repository look-up index
            DatabaseEntity ent = sdxRepoIndexDB.getEntity(doc.getId());
            if (ent == null) {
                String[] args = new String[2];
                args[0] = doc.getId();
                args[1] = this.getId();
                throw new SDXException(logger, SDXExceptionCode.ERROR_NO_DOC_EXISTS_DOCBASE, args, null);
            }
            ent.enableLogging(this.logger);
            String doctype = ent.getProperty(DOCUMENT_TYPE);
            if (Utilities.checkString(doctype)) {
                if (doctype.equalsIgnoreCase(Document.DOCTYPE_HTML))
                    doc = new HTMLDocument(doc.getId());
                else if (doctype.equalsIgnoreCase(Document.DOCTYPE_XML) || doctype.equalsIgnoreCase(Document.DOCTYPE_USER) || doctype.equalsIgnoreCase(Document.DOCTYPE_GROUP))
                    doc = new XMLDocument(doc.getId());
                //TODO:maybe in the future we need to add more checks for other types of parsable documents
            }

        }
        //now we know have a good doctype so we call our original method
        this.getDocument(doc, consumer);

        /*
        Repository repo = null;
        RepositoryConnection conn = null;

        try {
            //getting the repository in which the document resides
            repo = getRepositoryForDocument(doc);
            //getting a connection for the repository
            conn = repo.getConnection();
            //then call toSAX from the repository
            repo.toSAX(doc, consumer, conn);
        } catch (SDXException e) {
            throw e;
        } finally {
            //always close the connection
            if (repo != null && conn != null) repo.releaseConnection(conn);
        }
        */
    }

    /** Supplies the provided output stream with the requested document
     * @param doc	The document.
     * @param os	The output stream.
     * @throws SDXException
     */
    public void getDocument(Document doc, OutputStream os) throws SDXException {
        //ensuring we have valid objects
        super.getDocument(doc, os);
        Repository repo = null;
        RepositoryConnection conn = null;
        try {
            //getting the repository in which the document resides
            repo = getRepositoryForDocument(doc);
            //getting a connection for the repository
            conn = repo.getConnection();
            //then call toSAX from the repository
            repo.get(doc, os, conn);
        } catch (SDXException e) {
            throw e;
        } finally {
            //always close the connection
            if (repo != null && conn != null) repo.releaseConnection(conn);
        }
    }

    /** Provides the requested SDX document as an InputStream
     * @param doc	The document.
     * @return	An input stream.
     * @throws SDXException
     */
    public InputStream getDocument(Document doc) throws SDXException {
        //ensuring we have valid objects
        super.getDocument(doc);
        Repository repo = null;
        RepositoryConnection conn = null;
        InputStream is = null;

        try {

            //getting the repository in which the document resides
            repo = getRepositoryForDocument(doc);
            //getting a connection for the repository
            conn = repo.getConnection();
            //getting an InputStream from the repository with a null encoding
            //TODO?: should we support an encoding parameter in the future?-rbp
            is = repo.openStream(doc, null, conn);

        } finally {
            //always close the connection
            if (repo != null && conn != null) repo.releaseConnection(conn);
        }
        return is;
    }

    /**Gets the repository object in which a document resides,
     * document should not be null and have a valid id,
     * use Utilities.checkDocument before calling this method*/
    private Repository getRepositoryForDocument(Document doc) throws SDXException {

        //looking-up for the document in repository look-up index
        DatabaseEntity ent = sdxRepoIndexDB.getEntity(doc.getId());
        if (ent == null) {
            String[] args = new String[2];
            args[0] = doc.getId();
            args[1] = this.getId();
            throw new SDXException(logger, SDXExceptionCode.ERROR_NO_DOC_EXISTS_DOCBASE, args, null);
        }

        ent.enableLogging(this.logger);

        Repository repo;

        //getting the document's repository
        String repoId = ent.getProperty(REPO_PROPERTY);
        if (!Utilities.checkString(repoId)) {
            throw new SDXException(logger, SDXExceptionCode.ERROR_INVALID_REPO_ID, null, null);
        } else if (repositories != null && repositories.containsKey(repoId)) {
            repo = (Repository) repositories.get(repoId);
            return repo;
        } else if (defaultRepository != null) {
            repo = defaultRepository;
            return repo;
            //TODO : is it legitimate to return the default repository here ? The repository hasn't been clearly identified !-pb
            //i assume not, but would you like to see an error messge instead-rbp
            /*
			* Yes :-) I think that this method shouldn't return null but throw an
			* exception in EVERY case where the repository could not be actually
			* retrieved. IMHO, there are at least 2 reasons :
			* 1) this method is private
			* 2) a good implementation design would call this method in a try... catch block
			* I don't think it would be worth the value to distinguish between different causes
			* that could raise such exceptions ; we're at a very low level
			*/
        } else {
            //we have a serious problem because we somehow have no repositories
            String[] args = new String[1];
            args[0] = doc.getId();
            throw new SDXException(logger, SDXExceptionCode.ERROR_RETRIEVE_REPO_FOR_DOCUMENT, args, null);
        }
    }

    /**
     * Adds an indexable document to this base.
     *
     * @param originalDoc   The document to add and to index.
     * @param repository    The repository into which the document will be saved.
     * @param conn          The connection to the repository.
     */
   private void add(Document originalDoc, Repository repository, RepositoryConnection conn, IndexParameters params, boolean batchIndex, boolean isMaster) throws SDXException, SAXException {
        boolean handleConnection = false;

        try {
            //TODO : refactor connection management (in the calling method ? use a pool ?)         
            //some set-up for accessing the repository
            if (repository == null) repository = defaultRepository;
            handleConnection = (conn == null);
            /*
             * TODO : check if the *provided* connection if fit for the repository ?
             * the opportunity to default the repository could result in providing an irrelevant connection -pb
             */
            if (handleConnection) conn = repository.getConnection();

            Document docToStore = originalDoc; //By default, we store the original document
            boolean storeTransformed = false;

            /*
            * this flag replaces the tests on (doc instanceof IndexableDocument)
            * some instances of IndexableDocument (e.g. TransformedDocuments) are *not* to be indexed)            * 
            * TODO : find a better name and improve concept -pb
            */
            if (isMaster) {

                //Code safe here : null is assumed in case we can't get a transformed document
                IndexableDocument transformedDoc = ((IndexableDocument)originalDoc).getTransformedDocument();                

                //determining which document to store
                if (transformedDoc != null && !(repository instanceof URLRepository) && Utilities.checkString(transformedDoc.getId())) {
                    //modifying the original object's id and filename
                    originalDoc.setId("o_" + originalDoc.getId());
                    originalDoc.setPreferredFilename("o_" + originalDoc.getPreferredFilename());
                    docToStore = (Document)transformedDoc; //store the transformed
                    storeTransformed = true;
                }

                /*
                * Resolution of :
                * a) document id
                * b) attached documents info
                * c) search index document
                */
                org.apache.lucene.document.Document indexationDoc = getIndexationDocument((IndexableDocument)originalDoc, ((IndexableDocument)docToStore).getId());
                //TOTOpb : understand how ce could have gotten a valid id before this :-)

                /*
                * STEP 1 : adding the document's indexation to the search index
                */
                //TODO : possibly differ this step -pb            
                if (indexationDoc != null) {
                    addToSearchIndex(indexationDoc, batchIndex);
                }                

                /*
                * STEP 2 : adding the original document, if relevant
                */                
                if (storeTransformed && (params.isSaveOriginalDocument() && this.keepOriginalDocuments))
                {
                    //Notice that the original document look-up entry will be added as well
                    //Notice also that the original doc is down-casted to Document to avoid processes devoted to *actually* indexed docs
                    this.add((Document)originalDoc, repository, conn, null, batchIndex, false); //Recursive call -pb
                    
                    //TODO : uncomment when approved -pb
                    //adding the relation between the document and the original document
                    //addToRelationsIndex(docToStore, originalDoc, new String("original"), null, 1, batchIndex);
                }

                /*
                * STEP 3 : adding the attached documents, if relevant
                */          
                Enumeration attachedDocs = ((IndexableDocument)originalDoc).getAttachedDocuments();
                if (attachedDocs != null) {            
                    while (attachedDocs.hasMoreElements()) {
                        //getting the attached docucment
                        BinaryDocument attachedDoc = (BinaryDocument) attachedDocs.nextElement();
                        Utilities.checkDocument(logger, attachedDoc);
                        DatabaseEntity attachedDocLookupEntry = sdxRepoIndexDB.getEntity(attachedDoc.getId());
                        //if this attached document doesn't exist already, add it
                        if (attachedDocLookupEntry == null) {
                            //Notice that the attached document look-up entry will be added as well                            
                            this.add(attachedDoc, repository, conn, null, batchIndex, false); //Recursive call -pb                            

                            //TODO : uncomment when approved -pb
                            //adding the relation between the document and the attached document
                            //addToRelationsIndex(docToStore, attachedDoc, new String("attached"), null, 1, batchIndex);
                        }
                        else
                        {
                            /*
                            * WARNING :
                            * This implementation assumes that the attached document
                            * is the same than the one that is already stored.
                            * One can imagine that, for a *same* id, other properties
                            * (mimetype, data) could be different.
                            * We should be able to use IndexParamters and/or attachedDocLookupEntry
                            * to determine the different possible behaviours :
                            * always replace
                            * never replace (ignore)                            
                            * create new instance (in which case, the id management won't be easy)
                            * replace if "master document" (concept to be implemented)
                            * ...
                            *
                            * For now, we do nothing here since the attached document 
                            * is already present, probably (certainly ;-) owned by other documents -pb
                            */
                        }
                    }
                }
            }

            /*
            * STEP 4 : adding the document's entry to the look-up index
            */    
            if (storeTransformed)
                addToLookupIndex(docToStore, originalDoc.getId(), repository, batchIndex);
            else
                addToLookupIndex(docToStore, null, repository, batchIndex);

            /*
            * STEP 5 : adding the document itself
            */   
            repository.add(docToStore, conn);
           
        } catch (SDXException e) {
            throw e;
        } finally {
            //cleaning-up
            if (handleConnection && repository != null) repository.releaseConnection(conn);
            //TODO : delegate the release to the calling method -pb
        }
    }

    //TODO : possibly move this method to a LuceneLookupIndex class or to LuceneDatase -pb
    private void addToLookupIndex(Document doc, String originalId, Repository repository, boolean batchIndex)  throws SDXException {
        try {
            //creating a document entry in repository look-up index
            DatabaseEntity docLookupEntry = new DatabaseEntity(doc.getId());
            //TODO : check that no entry already exists for this id ? -pb
            docLookupEntry.enableLogging(this.logger);

            //saving the repository
            docLookupEntry.addProperty(REPO_PROPERTY, repository.getId());
            //saving the document type
            docLookupEntry.addProperty(DOCUMENT_TYPE, doc.getDocType());
            //saving the mime type
            if (Utilities.checkString(doc.getMimeType()))
                docLookupEntry.addProperty(MIMETYPE_PROPERTY, doc.getMimeType());
            //notice : reuse of the previous code regarding binary documents -pb
            else
                /*                  
                * TODO : does it have sense ?
                * it is assumed that every IndexableDocument has a mimetype
                * wouldn't it be consistent to also provide one for binary documents ? -pb
                */
                docLookupEntry.addProperty(MIMETYPE_PROPERTY, BinaryDocument.DEFAULT_MIMETYPE);
            /*
            * TODO : avoid this as much as possible 
            * because transformed documents *are* IndexableDocuments
            * that are *not* to be indexed
            * alternatives : add a toBeIndexed flag to Indexabledocument 
            * or use a isMaster flag like in add and delete -pb
            */
            if (doc instanceof IndexableDocument) { //TODO : this block is no more useful when relations are managed as is -pb                
                if (originalId != null)
                    docLookupEntry.addProperty(ORIGINAL_DOCUMENT_PROPERTY, originalId);           
                //processing attached documents
                Enumeration attachedDocs = ((IndexableDocument)doc).getAttachedDocuments();
                if (attachedDocs != null) {            
                    while (attachedDocs.hasMoreElements()) {
                        //getting the attached doc
                        BinaryDocument attachedDoc = (BinaryDocument) attachedDocs.nextElement();
                        //adding attached doc ids to the database entity for the parent document                
                        docLookupEntry.addProperty(ATTACHED_DOCS_PROPERTY, attachedDoc.getId());
                    }
                }
            } 
            //saving document entry in look-up index
            sdxRepoIndexDB.save(docLookupEntry);
            //if we don't have a batch index we optimize
            if (!batchIndex) sdxRepoIndexDB.optimize();
            //TODO : transfer the batch control here -pb
        } catch (SDXException e) {            
            throw e;
        }        
    }

    //TODO : possibly move this method to a LuceneRelationsIndex class or to LuceneDatase -pb
    private void addToRelationsIndex(Document docA, Document docB, String relationType, Integer relationPosition, int relationDirection, boolean batchIndex)  throws SDXException {    
        try {
            //creating a document entry in repository look-up index
            DatabaseEntity relationEntry = new DatabaseEntity(getRelationID(docA, docB, relationType, relationDirection));
            //TODO : check that no entry already exists for this id ? we have to be *very* confident into getRelationID() -pb
            relationEntry.enableLogging(this.logger);
            relationEntry.addProperty("docAID", docA.getId());
            //TODO : add repository when we have a Document.getRepository() method -pb
            relationEntry.addProperty("docBID", docB.getId());
            //TODO : add repository when we have a Document.getRepository() method -pb
            relationEntry.addProperty("type", relationType);
            relationEntry.addProperty("order", relationPosition.toString());
            relationEntry.addProperty("direction", String.valueOf(relationDirection));
            //saving relation entry in look-up index
            sdxRelationsIndexDB.save(relationEntry);

            logger.info("added relation : " + relationEntry.getId());

            //if we don't have a batch index we optimize
            if (!batchIndex) sdxRepoIndexDB.optimize();
            //TODO : transfer the batch control here -pb
        } catch (SDXException e) {            
            throw e;
        }        
    }

    //TODO : define a proper format -pb
    private String getRelationID(Document docA, Document docB, String relationType, int relationDirection)  throws SDXException {    
        StringBuffer ID = new StringBuffer();
        ID.append(docA.getId());
        switch (relationDirection) {
            case 0: //bidirectionnal ; TODO : use constants
                ID.append("<->");
            case 1 : //A to B ; TODO : use constants
                ID.append("->");
            case 2 : //A from B ; TODO : use constants 
                ID.append("<-");
            default :
                //TODO : throw exception
        }
        ID.append(docB.getId());
        ID.append(":");
        ID.append(relationType);
        return ID.toString();
    }

    //TODO : possibly move this method to the LuceneIndex class -pb
    private void addToSearchIndex(org.apache.lucene.document.Document indexationDoc, boolean batchIndex) throws SDXException {
        try {
            //writing the Lucene index document to the search index
            luceneSearchIndex.writeIndex(LuceneIndex.ACTION_ADD_DOCUMENT, indexationDoc, batchIndex);
            //TODO: use the below line once we are able make changes to lucene's code, but when?
            //writeIndex(ACTION_ADD_DOCUMENT, lDoc, field.getAnalyzer())
             //TODO : transfer the batch control here -pb
        } catch (SDXException e) {            
            throw e;
        }        
    }
    
    //TODO : possibly move this method to the LuceneIndex class -pb
    private org.apache.lucene.document.Document getIndexationDocument(IndexableDocument doc, String storeDocId) throws SDXException {

        //TODO : try... catch ; possibly return null -pb

        org.apache.lucene.document.Document lDoc = new org.apache.lucene.document.Document();

        //getting Lucene fields
        Enumeration fieldVals = doc.getFieldValues();
        while (fieldVals != null && fieldVals.hasMoreElements()) {
            Property prop = (Property) fieldVals.nextElement();
            //TODO: what do we do with non defined fields??? For now we will have a NPE...maybe just index and tokenize no store!
            fr.gouv.culture.sdx.search.lucene.Field sdxF = fieldsDef.getField(prop.getName());
            if (sdxF != null) {
                org.apache.lucene.document.Field f = sdxF.getLuceneField(prop.getValue());
                lDoc.add(f);
            }
            // else  return null //TODOLogging:log.warn or info? the unknown fields
        }

        //if we dont have a document id after the parse or for the attached document, we have a problem and need throw an error
        Utilities.checkDocument(logger, doc);

        //adding system metadata
        Field fId = null;
        if (Utilities.checkString(storeDocId))
            fId = Field.Keyword(INTERNAL_FIELD_NAME_SDXDOCID, storeDocId);
        if (fId != null)
            lDoc.add(fId);

        Field fId2 = Field.Keyword(INTERNAL_FIELD_NAME_SDXALL, INTERNAL_FIELD_VALUE_SDXALL);
        if (fId2 != null)
            lDoc.add(fId2);

        Field fId3 = null;
        String docbaseId = this.getId();
        if (Utilities.checkString(docbaseId))
            fId3 = Field.Keyword(INTERNAL_FIELD_NAME_SDXDBID, docbaseId);
        if (fId3 != null)
            lDoc.add(fId3);

        //Warning : what if this DB is shared between applications ?
        Field fId4 = null;
        String appId = Utilities.getStringFromHashtable(Application.APPLICATION_ID, props);
        if (Utilities.checkString(appId))
            fId4 = Field.Keyword(INTERNAL_FIELD_NAME_SDXAPPID, appId);
        if (fId4 != null)
            lDoc.add(fId4);
        Field fId5 = null;
        String doctype = doc.getDocType();
        if (Utilities.checkString(doctype))
            fId5 = Field.Keyword(INTERNAL_FIELD_NAME_SDXDOCTYPE, doctype);
        if (fId5 != null)
            lDoc.add(fId5);
        return lDoc;
    }

    /**Supplies the mimeType for the the document if it exists
     * otherwise the binary document default mimetype is returned
     * @param doc	The document.
     */
    //TODO : I can't understand a default result down here. Either the mimetype is actually retrieved
    //either it's not -pb
    public String getMimeType(Document doc) {
        DatabaseEntity ent = null;
        try {
            ent = sdxRepoIndexDB.getEntity(doc.getId());
        } catch (SDXException e) {
            return BinaryDocument.DEFAULT_MIMETYPE;
        }
        if (ent == null)
            return BinaryDocument.DEFAULT_MIMETYPE;

        ent.enableLogging(this.logger);
        String mimeType = ent.getProperty(MIMETYPE_PROPERTY);
        if (!Utilities.checkString(mimeType))
            return BinaryDocument.DEFAULT_MIMETYPE;
        else
            return mimeType;
    }

}
